"use strict";Object.defineProperty(exports, "__esModule", {value: true});/**
 * Pokemon Showdown log viewer
 *
 * by Zarel
 */

var _fs = require('../../............................................lib-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist-dist/fs');

class LogReaderRoom {
	
	constructor(roomid) {
		this.roomid = roomid;
	}

	async listMonths() {
		try {
			const listing = await _fs.FS.call(void 0, `logs/chat/${this.roomid}`).readdir();
			return listing.filter(file => /^[0-9][0-9][0-9][0-9]-[0-9][0-9]$/.test(file));
		} catch (err) {
			return [];
		}
	}

	async listDays(month) {
		try {
			const listing = await _fs.FS.call(void 0, `logs/chat/${this.roomid}/${month}`).readdir();
			return listing.filter(file => /\.txt$/.test(file)).map(file => file.slice(0, -4));
		} catch (err) {
			return [];
		}
	}

	async getLog(day) {
		const month = LogReader.getMonth(day);
		const log = _fs.FS.call(void 0, `logs/chat/${this.roomid}/${month}/${day}.txt`);
		if (!await log.exists()) return null;
		return log.createReadStream();
	}
}

const DAY = 24 * 60 * 60 * 1000;

const LogReader = new class {
	async get(roomid) {
		if (!await _fs.FS.call(void 0, `logs/chat/${roomid}`).exists()) return null;
		return new LogReaderRoom(roomid);
	}

	async list() {
		const listing = await _fs.FS.call(void 0, `logs/chat`).readdir();
		return listing.filter(file => /^[a-z0-9-]+$/.test(file));
	}

	getMonth(day) {
		return day.slice(0, 7);
	}
	nextDay(day) {
		const nextDay = new Date(new Date(day).getTime() + DAY);
		return nextDay.toISOString().slice(0, 10);
	}
	prevDay(day) {
		const prevDay = new Date(new Date(day).getTime() - DAY);
		return prevDay.toISOString().slice(0, 10);
	}
	nextMonth(month) {
		const nextMonth = new Date(new Date(`${month}-15`).getTime() + 30 * DAY);
		return nextMonth.toISOString().slice(0, 7);
	}
	prevMonth(month) {
		const prevMonth = new Date(new Date(`${month}-15`).getTime() - 30 * DAY);
		return prevMonth.toISOString().slice(0, 7);
	}
};

const LogViewer = new class {
	async day(roomid, day, opts) {
		const month = LogReader.getMonth(day);
		let buf = `<div class="pad"><p>` +
			`<a roomid="view-chatlog">← All logs</a> / ` +
			`<a roomid="view-chatlog-${roomid}">${roomid}</a> /  ` +
			`<a roomid="view-chatlog-${roomid}--${month}">${month}</a> / ` +
			`<strong>${day}</strong></p>`;

		const roomLog = await LogReader.get(roomid);
		if (!roomLog) {
			buf += `<p class="message-error">Room "${roomid}" doesn't exist</p></div>`;
			return this.linkify(buf);
		}

		buf += `<p><a roomid="view-chatlog-${roomid}--${LogReader.prevDay(day)}">⬆️ Earlier</a></p><div>`;

		const stream = await roomLog.getLog(day);
		if (!stream) {
			buf += `<p class="message-error">Room "${roomid}" doesn't have logs for ${day}</p>`;
		} else {
			let line;
			while ((line = await stream.readLine()) !== null) {
				buf += this.renderLine(line, opts);
			}
		}

		buf += `</div><p><a roomid="view-chatlog-${roomid}--${LogReader.nextDay(day)}">⬇️ Later</a></p>`;

		buf += `</div>`;
		return this.linkify(buf);
	}
	renderLine(fullLine, opts) {
		const timestamp = fullLine.slice(0, opts ? 8 : 5);
		const line = fullLine.charAt(9) === '|' ? fullLine.slice(10) : '|' + fullLine.slice(9);
		if (opts !== 'all' && (
			line.startsWith(`userstats|`) ||
			line.startsWith('J|') || line.startsWith('L|') || line.startsWith('N|')
		)) return ``;

		const cmd = line.slice(0, line.indexOf('|'));
		switch (cmd) {
		case 'c': {
			const [, name, message] = Chat.splitFirst(line, '|', 2);
			if (name.length <= 1) {
				return `<div class="chat"><small>[${timestamp}] </small><em>${Chat.formatText(message)}</em></div>`;
			}
			return `<div class="chat"><small>[${timestamp}] </small><strong><small>${name.charAt(0)}</small>${name.slice(1)}:</strong> <em>${Chat.formatText(message)}</em></div>`;
		}
		case 'html': {
			const [, html] = Chat.splitFirst(line, '|', 1);
			return `<div class="notice">${html}</div>`;
		}
		case '':
			return `<div class="chat"><small>[${timestamp}] </small>${Chat.escapeHTML(line.slice(1))}</div>`;
		default:
			return `<div class="chat"><small>[${timestamp}] </small>${'|' + Chat.escapeHTML(line)}</div>`;
		}
	}
	async month(roomid, month) {
		let buf = `<div class="pad"><p>` +
			`<a roomid="view-chatlog">← All logs</a> / ` +
			`<a roomid="view-chatlog-${roomid}">${roomid}</a> / ` +
			`<strong>${month}</strong></p>`;

		const roomLog = await LogReader.get(roomid);
		if (!roomLog) {
			buf += `<p class="message-error">Room "${roomid}" doesn't exist</p></div>`;
			return this.linkify(buf);
		}

		buf += `<p><a roomid="view-chatlog-${roomid}--${LogReader.prevMonth(month)}">⬆️ Earlier</a></p>`;

		const days = await roomLog.listDays(month);
		if (!days.length) {
			buf += `<p class="message-error">Room "${roomid}" doesn't have logs in ${month}</p></div>`;
			return this.linkify(buf);
		} else {
			for (const day of days) {
				buf += `<p>- <a roomid="view-chatlog-${roomid}--${day}">${day}</a></p>`;
			}
		}

		buf += `<p><a roomid="view-chatlog-${roomid}--${LogReader.nextMonth(month)}">⬇️ Later</a></p>`;

		buf += `</div>`;
		return this.linkify(buf);
	}
	async room(roomid) {
		let buf = `<div class="pad"><p>` +
			`<a roomid="view-chatlog">← All logs</a> / ` +
			`<strong>${roomid}</strong></p>`;

		const roomLog = await LogReader.get(roomid);
		if (!roomLog) {
			buf += `<p class="message-error">Room "${roomid}" doesn't exist</p></div>`;
			return this.linkify(buf);
		}

		const months = await roomLog.listMonths();
		if (!months.length) {
			buf += `<p class="message-error">Room "${roomid}" doesn't have logs</p></div>`;
			return this.linkify(buf);
		}

		for (const month of months) {
			buf += `<p>- <a roomid="view-chatlog-${roomid}--${month}">${month}</a></p>`;
		}
		buf += `</div>`;
		return this.linkify(buf);
	}
	error(message) {
		return `<div class="pad"><p class="message-error">${message}</p></div>`;
	}
	linkify(buf) {
		return buf.replace(/<a roomid="/g, `<a target="replace" href="/`);
	}
	list() {
		return this.error(`Unimplemented.`);
	}
};

const accessLog = _fs.FS.call(void 0, `logs/chatlog-access.txt`).createAppendStream();

 const pages = {
	async chatlog(args, user, connection) {
		if (!user.named) return Rooms.RETRY_AFTER_LOGIN;
		if (!user.trusted) {
			return LogViewer.error("Access denied");
		}

		const [roomid, date, opts] = args.join('-').split('--') ;

		if (!roomid) {
			this.title = '[Logs]';
			return LogViewer.list();
		}

		// permission check
		const room = Rooms.get(roomid);
		if (roomid.startsWith('spl') && roomid !== 'splatoon' && !user.can('rangeban')) {
			return LogViewer.error("SPL team discussions are super secret.");
		}
		if (roomid.startsWith('wcop') && !user.can('rangeban')) {
			return LogViewer.error("WCOP team discussions are super secret.");
		}
		if (room) {
			if (!room.checkModjoin(user) && !user.can('bypassall')) {
				return LogViewer.error("Access denied");
			}
			if (!this.can('lock', null, room )) return;
		} else {
			if (!this.can('lock')) return;
		}

		void accessLog.writeLine(`${user.id}: <${roomid}> ${date}`);

		this.title = '[Logs] ' + roomid;
		if (date && date.length === 10) {
			return LogViewer.day(roomid, date, opts);
		}
		if (date) {
			return LogViewer.month(roomid, date);
		}
		return LogViewer.room(roomid);
	},
}; exports.pages = pages;
